from lru import LRU
from functools import wraps
import time
from coc.utils import find
import inspect
import logging
import re

log = logging.getLogger(__name__)


def validate_tag(string):
    # Legal clan tags only have these characters:
    # Numbers: 0, 2, 8, 9
    # Letters: P, Y, L, Q, G, R, J, C, U, V
    if not isinstance(string, str):
        return False

    match = re.match("(?P<tag>^\s*#?[PYLQGRJCUV0289]+\s*$)|(?P<location>\d{1,10})", string)

    if match.group('tag'):
        return True
    if match.group('location'):
        return True

    return False


def find_key(args, kwargs):
    if args:
        key = find(lambda x: validate_tag(x), args)
        if key:
            return key

    for k, v in kwargs.items():
        key = validate_tag(v)
        if key:
            return key

    return None


def _wrap_coroutine(result):
    async def new_coro():
        return result[0]
    return new_coro()


def _wrap_store_coro(cache, key, coro):
    async def fctn():
        value = await coro
        cache[key] = (value, time.monotonic())
        return value
    return fctn()


class Cache:
    def __init__(self, max_size=128, ttl=None):
        self.__lru_instance = LRU(max_size)
        self.__ttl = ttl

    def __call__(self, *args, **kwargs):
        self.check_expiry()

    def check_expiry(self):
        if not self.__ttl:
            return

        current_time = time.monotonic()
        to_delete = [k for (k, (v, t)) in self.__lru_instance.items() if current_time > (t + self.__ttl)]
        for k in to_delete:
            log.debug('Removed item with key %s and TTL %s seconds from cache.', k, self.__ttl)
            del self.__lru_instance[k]

    def get_cache(self):
        def deco(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                self.check_expiry()
                key = find_key(args, kwargs)
                cache = kwargs.pop('cache', False)
                fetch = kwargs.pop('fetch', True)

                if not key:
                    return func(*args, **kwargs)
                try:
                    if cache:
                        data = self.__lru_instance[key]
                    else:
                        if fetch:
                            data = func(*args, **kwargs)
                            return _wrap_store_coro(self.__lru_instance,
                                                    key, data)
                        else:
                            return None

                except KeyError:
                    if fetch:
                        data = func(*args, **kwargs)
                    else:
                        return None

                    if inspect.isasyncgen(data):
                        return data

                    if inspect.isawaitable(data):
                        return _wrap_store_coro(self.__lru_instance,
                                                key, data)

                    self.__lru_instance[key] = (data, time.monotonic())
                    return data

                else:
                    log.debug('Using object cached at %s', data[1])
                    if inspect.iscoroutinefunction(func):
                        return _wrap_coroutine(data)

                    return data

            return wrapper
        return deco

    def get(self, key):
        try:
            return self.__lru_instance[key]
        except KeyError:
            return None

    def add(self, key, value):
        self.__lru_instance[key] = (value, time.monotonic())




# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 mathsman5133

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from datetime import datetime
from operator import attrgetter
from itertools import chain

from .utils import get, from_timestamp


def try_enum(_class, data):
    if data is None:
        return None
    return _class(data=data)


class Clan:
    """Represents the most stripped down version of clan info.
    All other clan classes inherit this.

    Attributes
    ------------
    tag:
        :class:`str` - The clan tag
    name:
        :class:`str` - The clan name
    """
    __slots__ = ('tag', 'name', '_data')

    def __init__(self, *, data):
        self._data = data
        self.tag = data.get('tag')
        self.name = data.get('name')

    @property
    def badge(self):
        """:class:`Badge` - The clan badges"""
        return try_enum(Badge, self._data.get('badgeUrls'))

    def __str__(self):
        return self.name


class BasicClan(Clan):
    """Represents a Basic Clan that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits :class:`Clan`, and thus all attributes
    of :class:`Clan` can be expected to be present.

    Attributes
    -----------
    level:
        :class:`int` - The clan level.
    points:
        :class:`int` - The clan trophy points.
    versus_points:
        :class:`int` - The clan versus trophy points.
    member_count:
        :class:`int` - The member count of the clan
    rank:
        :class:`int` - The clan rank for it's location this season
    previous_rank:
        :class:`int` - The clan rank for it's location in the previous season
    """
    __slots__ = ('level', 'points', 'versus_points',
                 'member_count', 'rank', 'previous_rank')

    def __init__(self, *, data):
        super().__init__(data=data)

        self.level = data.get('clanLevel')
        self.points = data.get('clanPoints')
        self.versus_points = data.get('clanVersusPoints')
        self.member_count = data.get('members')
        self.rank = data.get('rank')
        self.previous_rank = data.get('previous_rank')

    @property
    def location(self):
        """:class:`Location` - The clan's location"""
        return try_enum(Location, self._data.get('location'))


class SearchClan(BasicClan):
    """Represents a Searched Clan that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits both :class:`Clan` and :class:`BasicClan`,
    and thus all attributes of these classes can be expected to be present.

    Attributes
    -----------
    type:
        :class:`str` - The clan type: open, closed, invite-only etc.
    required_trophies:
        :class:`int` - The required trophies to join
    war_frequency:
        :class:`str` - The war frequency of the clan
    war_win_streak:
        :class:`int` - The current war win streak of the clan
    war_wins:
        :class:`int` - The total war wins of the clan
    war_ties:
        :class:`int` - The total war ties of the clan
    war_losses:
        :class:`int` - The total war losses of the clan
    public_war_log:
        :class:`bool` - Indicates whether the war log is public
    description:
        :class:`str` - The clan description
    """
    __slots__ = ('type', 'required_trophies', 'war_frequency', 'war_win_streak',
                 'war_wins', 'war_ties', 'war_losses', 'public_war_log',
                 'description', '_http')

    def __init__(self, *, data, http):
        super().__init__(data=data)

        self._http = http
        self.type = data.get('type')
        self.required_trophies = data.get('requiredTrophies')
        self.war_frequency = data.get('warFrequency')
        self.war_win_streak = data.get('warWinStreak')
        self.war_wins = data.get('warWins')
        self.war_ties = data.get('warTies')
        self.war_losses = data.get('warLosses')
        self.public_war_log = data.get('isWarLogPublic')
        self.description = data.get('description', '')

    @property
    def _members(self):
        """|iter|

        Returns an iterable of :class:`BasicPlayer`: A list of clan members.
        """
        return iter(BasicPlayer(mdata, self) for mdata in self._data.get('memberList', []))

    @property
    def members(self):
        """List[:class:`BasicPlayer`]: A list of clan members"""
        return list(self._members)

    @property
    def member_dict(self, attr='tag'):
        """Dict{attr: :class:`BasicPlayer`}: A dict of clan members by tag.

        Pass in an attribute of :class:`BasicPlayer` to get that attribute as the key
        """
        return {getattr(m, attr): m for m in self._members}

    def get_member(self, **attrs):
        """Returns the first :class:`BasicPlayer` that meets the attributes passed

        This will return the first member matching the attributes passed.

        An example of this looks like:

        .. code-block:: python3

            member = SearchClan.get_member(tag='tag')

        This search implements the :func:`coc.utils.get` function
        """
        return get(self._members, **attrs)

    async def get_detailed_members(self, cache=False, fetch=True):
        """Get detailed player information for every player in the clan.
        This will return an AsyncIterator of :class:`SearchPlayer`.

        Example
        --------

        .. code-block:: python3

            clan = client.get_clan('tag')

            async for player in clan.get_detailed_members(cache=True):
                print(player.name)

        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned
        :return: AsyncIterator of :class:`SearchPlayer`
        """
        for n in self._members:
            yield self._http.client.get_player(n.tag, cache=cache, fetch=fetch)


class WarClan(Clan):
    """Represents a War Clan that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits :class:`Clan`, and thus all attributes
    of :class:`Clan` can be expected to be present.

    Attributes
    -----------
    attack_count:
        :class:`int` - Number of attacks by clan this war
    stars:
        :class:`int` - Number of stars by clan this war
    destruction:
        :class:`float` - Destruction as a percentage
    exp_earned:
        :class:`int` - Total XP earned by clan this war
    attacks_used:
        :class:`int` - Total attacks used by clan this war
    max_stars:
        :class:`int` - Total possible stars achievable
    """
    __slots__ = ('_war', 'level',
                 'attack_count', 'stars', 'destruction', 'exp_earned',
                 'attacks_used', 'total_attacks', 'max_stars')

    def __init__(self, *, data, war):
        super(WarClan, self).__init__(data=data)

        self._war = war
        self.level = data.get('clanLevel')
        self.attack_count = data.get('attacks')
        self.stars = data.get('stars')
        self.destruction = data.get('destructionPercentage')
        self.exp_earned = data.get('expEarned')

        self.attacks_used = data.get('attacks')
        self.total_attacks = self._war.team_size * 2
        self.stars = data.get('stars')
        self.max_stars = self._war.team_size * 3
            
    @property
    def _members(self):
        """|iter|

        Returns an iterable of :class:`WarMember`: all clan members in war.
        """
        return iter(WarMember(data=mdata, war=self._war, clan=self)
                    for mdata in self._data.get('members', []))

    @property
    def members(self):
        """List[:class:`WarMember`]: List of all clan members in war"""
        return list(self._members)

    @property
    def members_dict(self, attr='tag'):
        """Dict{attr: :class:`WarMember`}: A dict of clan members in war by tag.

        Pass in an attribute of :class:`WarMember` to get that attribute as the key.
        """
        return {getattr(m, attr): m for m in self._members}

    @property
    def _attacks(self):
        """|iter|

        Returns an iterable of :class:`WarAttack`: all attacks used by the clan this war.
        """
        return chain.from_iterable(iter(n._attacks for n in self._members))

    @property
    def attacks(self):
        """List[:class:`WarAttack`]: List of all attacks used this war."""
        return list(self._attacks)

    @property
    def _defenses(self):
        """|iter|

        Returns an iterable of :class:`WarAttack`: all defenses by clan members this war.
        """
        return chain.from_iterable(iter(n._defenses for n in self._members))

    @property
    def defenses(self):
        """List[:class:`WarAttack`]: List of all defenses by clan members this war."""
        return list(self._defenses)


class Player:
    """Represents the most stripped down version of a player.
    All other player classes inherit this.


    Attributes
    ------------
    tag:
        :class:`str` - The clan tag
    name:
        :class:`str` - The clan name
    """
    __slots__ = ('name', 'tag', '_data')

    def __init__(self, data):
        self._data = data
        self.name = data['name']
        self.tag = data.get('tag')

    def __str__(self):
        return self.name


class BasicPlayer(Player):
    """Represents a Basic Player that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits :class:`Player`, and thus all attributes
    of :class:`Player` can be expected to be present.

    Attributes
    -----------
    clan:
        :class:`Basic Clan` - The clan the member belongs to. May be ``None``
    level:
        :class:`int` - The player level.
    trophies:
        :class:`int` - The player's trophy count.
    versus_trophies:
        :class:`int` - The player's versus trophy count.
    clan_rank:
        :class:`int` - The members clan rank
    clan_previous_rank
        :class:`int` - The members clan rank last season
    league_rank:
        :class:`int` - The player's current rank in their league for this season
    donations:
        :class:`int` - The members current donation count
    received:
        :class:`int` - The member's current donation received count
    attack_wins:
        :class:`int` - The players current attack wins for this season
    defense_wins:
        :class:`int` - The players current defense wins for this season
    """
    __slots__ = ('clan', 'level', 'trophies', 'versus_trophies',
                 'clan_rank', 'clan_previous_rank', 'league_rank', 'donations',
                 'received', 'attack_wins', 'defense_wins')

    def __init__(self, data, clan=None):
        super(BasicPlayer, self).__init__(data)

        self.clan = clan
        self.level = data.get('expLevel')
        self.trophies = data.get('trophies')
        self.versus_trophies = data.get('versusTrophies')
        self.clan_rank = data.get('clanRank')
        self.clan_previous_rank = data.get('clanRank')
        self.league_rank = data.get('rank')
        self.donations = data.get('donations')
        self.received = data.get('donationsReceived')
        self.attack_wins = data.get('attackWins')
        self.defense_wins = data.get('defenseWins')

        if not self.clan:
            cdata = data.get('clan')
            if cdata:
                self.clan = BasicClan(data=cdata)

    @property
    def league(self):
        """:class:`League`: The player's current league."""
        return try_enum(League, self._data.get('league'))

    @property
    def role(self):
        """:class:`str`: The members role in the clan - member, elder, etc."""
        role = self._data.get('role')
        if role == 'admin':
            return 'elder'
        return role


class WarMember(Player):
    """Represents a War Member that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits :class:`Player`, and thus all attributes
    of :class:`Player` can be expected to be present.


    Attributes
    -----------
    town_hall:
        :class:`int` - The members TH level
    map_position:
        :class:`int` - The members map position this war
    war:
        :class:`War` - The war this member belongs to
    clan:
        :class:`WarClan` - The war clan this member belongs to.
    """
    __slots__ = ('town_hall', 'map_position', 'war', 'clan')

    def __init__(self, data, war, clan):
        super(WarMember, self).__init__(data)

        self.war = war
        self.clan = clan

        self.town_hall = data.get('townHallLevel')
        self.map_position = data.get('mapPosition')

    @property
    def _attacks(self):
        """|iter|

        Returns an iterable of :class:`WarAttack`: the member's attacks this war"""
        return iter(WarAttack(data=adata, war=self.war, member=self)
                    for adata in self._data.get('attacks', []))

    @property
    def attacks(self):
        """List[:class:`WarAttack`]: The member's attacks this war. Could be an empty list"""
        return list(self._attacks)

    @property
    def _defenses(self):
        """|iter|

        Returns an iterable of :class:`WarAttack`: the member's defenses this war"""
        return filter(lambda o: o.defender_tag == self.tag, self.war.opponent._attacks)

    @property
    def defenses(self):
        """List[:class:`WarAttack`]: The member's defenses this war. Could be an empty list"""
        return list(self._defenses)


class SearchPlayer(BasicPlayer):
    """Represents a Searched Player that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    This class inherits both :class:`Player` and :class:`BasicPlayer`,
    and thus all attributes of these classes can be expected to be present

    Attributes
    -----------
    best_trophies:
        :class:`int` - The players top trophy count
    best_versus_trophies:
        :class:`int` - The players top versus trophy count
    war_stars:
        :class:`int` - The players war star count
    town_hall:
        :class:`int` - The players TH level
    builder_hall:
        :class:`int` - The players BH level
    versus_attacks_wins:
        :class:`int` - The players total BH wins
    """
    __slots__ = ('best_trophies', 'war_stars', 'town_hall',
                 'builder_hall', 'best_versus_trophies', 'versus_attacks_wins')

    def __init__(self, *, data):
        super(SearchPlayer, self).__init__(data=data)

        self.clan = try_enum(Clan, data.get('clan'))
        self.best_trophies = data.get('bestTrophies')
        self.war_stars = data.get('warStars')
        self.town_hall = data.get('townHallLevel')
        self.builder_hall = data.get('builderHallLevel')
        self.best_versus_trophies = data.get('bestVersusTrophies')
        self.versus_attacks_wins = data.get('versusBattleWins')

    @property
    def _achievements(self):
        """|iter|

        Returns an iterable of :class:`Achievement`: the player's achievements."""
        return iter(Achievement(data=adata, player=self)
                    for adata in self._data.get('achievements', []))

    @property
    def achievements(self):
        """List[:class:`Achievement`]: List of the player's achievements"""
        return list(self._achievements)

    @property
    def troops(self):
        """List[:class:`Troop`]: List of the player's troops"""
        return [Spell(data=sdata, player=self)
                for sdata in self._data.get('troops', [])]

    @property
    def heroes(self):
        """List[:class:`Hero`]: List of the player's heroes"""
        return [Hero(data=hdata, player=self)
                for hdata in self._data.get('heroes', [])]

    @property
    def spells(self):
        """List[:class:`Spell`]: List of the player's spells"""
        return [Spell(data=sdata, player=self)
                for sdata in self._data.get('spells', [])]
            
    @property
    def achievements_dict(self, attr='name'):
        """Dict{name: :class:`Achievement`}: A dict of achievements by name.

        Pass in an attribute of :class:`Achievement` to get that attribute as the key
        """
        return {getattr(m, attr): m for m in self._achievements}

    @property
    def troops_dict(self, attr='name'):
        """Dict{name: :class:`Troop`}: A dict of troops by name.

        Pass in an attribute of :class:`Troop` to get that attribute as the key
        """
        return {getattr(m, attr): m for m in self.troops}

    @property
    def heroes_dict(self, attr='name'):
        """Dict{name: :class:`Hero`}: A dict of heroes by name.

        Pass in an attribute of :class:`Hero` to get that attribute as the key
        """
        return {getattr(m, attr): m for m in self.heroes}

    @property
    def spells_dict(self, attr='name'):
        """Dict{name: :class:`Spell`}: A dict of spells by name.

        Pass in an attribute of :class:`Spell` to get that attribute as the key
        """
        return {getattr(m, attr): m for m in self.spells}


class BaseWar:
    """Represents the most basic Clash of Clans War

    Attributes
    -----------
    team_size:
        :class:`int` - The number of players per clan in war
    """
    __slots__ = ('team_size', '_data')

    def __init__(self, *, data):
        self._data = data
        self.team_size = data.get('teamSize')

    @property
    def clan(self):
        """:class:`WarClan`: The offensive clan"""
        clan = self._data.get('clan', {})
        if 'tag' in clan:
            # at the moment, if the clan is in notInWar, the API returns
            # 'clan' and 'opponent' as dicts containing only badge urls of
            # no specific clan. very strange

            return WarClan(data=clan, war=self)

    @property
    def opponent(self):
        """:class:`WarClan`: The opposition clan"""
        opponent = self._data.get('opponent', {})
        if 'tag' in opponent:
            # same issue as clan
            return WarClan(data=opponent, war=self)
        

class WarLog(BaseWar):
    """Represents a Clash of Clans War Log Entry

    This class inherits :class:`BaseWar`, and thus all attributes
    of :class:`BaseWar` can be expected to be present.

    Attributes
    -----------
    result:
        :class:`str` - The result of the war - `win` or `loss`
    end_time:
        :class:`Timestamp` - The end time of the war as a Timestamp object
    """
    __slots__ = ('result', 'end_time')

    def __init__(self, *, data):
        self.result = data.get('result')
        self.end_time = try_enum(Timestamp, data.get('endTime'))
        super(WarLog, self).__init__(data=data)


class CurrentWar(BaseWar):
    """Represents a Current Clash of Clans War

    This class inherits :class:`BaseWar`, and thus all attributes
    of :class:`BaseWar` can be expected to be present.

    Attributes
    -----------
    state:
        :class:`str` - The clan's current war state
    preparation_start_time:
        :class:`Timestamp` - The start time of preparation day as a Timestamp object
    start_time:
        :class:`Timestamp` - The start time of battle day as a Timestamp object
    end_time:
        :class:`Timestamp` - The end time of battle day as a Timestamp object
    """
    __slots__ = ('state', 'preparation_start_time',
                 'start_time', 'end_time')

    def __init__(self, *, data):
        self.state = data.get('state')
        self.preparation_start_time = try_enum(Timestamp, data.get('preparationStartTime'))
        self.start_time = try_enum(Timestamp, data.get('startTime'))
        self.end_time = try_enum(Timestamp, data.get('endTime'))

        super(CurrentWar, self).__init__(data=data)

    @property
    def _attacks(self):
        """|iter|

        Returns an iterable of :class:`WarAttack`: all attacks this war"""
        return chain(self.clan._attacks, self.opponent._attacks)

    @property
    def attacks(self):
        """List[:class:`WarAttack`]: A list of all attacks this war"""
        a = list(self._attacks)
        a.sort(key=attrgetter('order'))
        return a

    @property
    def _members(self):
        """|iter|

        Returns an iterable of :class:`WarMember`: all members this war"""
        return chain(self.clan._members, self.opponent._members)

    @property
    def members(self):
        """List[:class:`WarMember`]: A list of all members this war"""
        return list(self._members)

    def get_member(self, **attrs):
        """Returns the first :class:`WarMember` that meets the attributes passed

        This will return the first member matching the attributes passed.

        An example of this looks like:

        .. code-block:: python3

            member = CurrentWar.get_member(tag='tag')

        This search implements the :func:`coc.utils.get` function
        """
        return get(self._members, **attrs)


class Achievement:
    """Represents a Clash of Clans Achievement.


    Attributes
    -----------
    player:
        :class:`SearchPlayer` - The player this achievement is assosiated with
    name:
        :class:`str` - The name of the achievement
    stars:
        :class:`int` - The current stars achieved for the achievement
    value:
        :class:`int` - The number of X things attained for this achievement
    target:
        :class:`int` - The number of X things required to complete this achievement
    info:
        :class:`str` - Information regarding the achievement
    completion_info:
        :class:`str` - Information regarding completion of the achievement
    village:
        :class:`str` - Either `home` or `builderBase`
    """

    __slots__ = ('player', 'name', 'stars', 'value', 'target',
                 'info', 'completion_info', 'village', '_data')

    def __str__(self):
        return self.name

    def __init__(self, *, data, player):
        self._data = data

        self.player = player
        self.name = data['name']
        self.stars = data.get('stars')
        self.value = data['value']
        self.target = data['target']
        self.info = data['info']
        self.completion_info = data.get('completionInfo')
        self.village = data['village']

    @property
    def is_builder_base(self):
        """:class:`bool`: Helper property to tell you if the achievement belongs to the builder base"""
        return self.village == 'builderBase'

    @property
    def is_home_base(self):
        """:class:`bool`: Helper property to tell you if the achievement belongs to the home base"""
        return self.village == 'home'

    @property
    def is_completed(self):
        """:class:`bool`: Indicates whether the achievement is completed (3 stars achieved)"""
        return self.stars == 3


class Troop:
    """Represents a Clash of Clans Troop.

    Attributes
    -----------
    player:
        :class:`SearchPlayer` - player this troop is assosiated with
    name:
        :class:`str` - The name of the troop
    level:
        :class:`int` - The level of the troop
    max_level:
        :class:`int` - The overall max level of the troop, excluding townhall limitations
    village:
        :class:`str` - Either `home` or `builderBase`
    """
    __slots__ = ('player', 'name', 'level',
                 'max_level', 'village', '_data')

    def __str__(self):
        return self.name

    def __init__(self, *, data, player):
        self._data = data

        self.player = player
        self.name = data['name']
        self.level = data['level']
        self.max_level = data['maxLevel']
        self.village = data['village']

    @property
    def is_max(self):
        """:class:`bool`: Helper property to tell you if the troop is the max level"""
        return self.max_level == self.level

    @property
    def is_builder_base(self):
        """:class:`bool`: Helper property to tell you if the troop belongs to the builder base"""
        return self.village == 'builderBase'

    @property
    def is_home_base(self):
        """:class:`bool`: Helper property to tell you if the troop belongs to the home base"""
        return self.village == 'home'


class Hero:
    """Represents a Clash of Clans Hero.

    Attributes
    -----------
    player:
        :class:`SearchPlayer` - The player this hero is assosiated with
    name:
        :class:`str` - The name of the hero
    level:
        :class:`int` - The level of the hero
    max_level:
        :class:`int` - The overall max level of the hero, excluding townhall limitations
    village:
        :class:`str` - Either `home` or `builderBase`
    """
    __slots__ = ('player', 'name', 'level',
                 'max_level', 'village', '_data')

    def __str__(self):
        return self.name

    def __init__(self, *, data, player):
        self._data = data

        self.player = player
        self.name = data['name']
        self.level = data['level']
        self.max_level = data['maxLevel']
        self.village = data['village']

    @property
    def is_max(self):
        """:class:`bool`: Helper property to tell you if the hero is the max level"""
        return self.level == self.max_level

    @property
    def is_builder_base(self):
        """:class:`bool`: Helper property to tell you if the hero belongs to the builder base"""
        return self.village == 'builderBase'

    @property
    def is_home_base(self):
        """:class:`bool`: Helper property to tell you if the hero belongs to the home base"""
        return self.village == 'home'


class Spell:
    """Represents a Clash of Clans Spell.

    Attributes
    -----------
    player:
        :class:`SearchPlayer` - The player this spell is assosiated with
    name:
        :class:`str` - The name of the spell
    level:
        :class:`int` - The level of the spell
    max_level:
        :class:`int` - The overall max level of the spell, excluding townhall limitations
    village:
        :class:`str` - Either `home` or `builderBase`
    """
    __slots__ = ('player', 'name', 'level',
                 'max_level', 'village', '_data')

    def __str__(self):
        return self.name

    def __init__(self, *, data, player):
        self._data = data

        self.player = player
        self.name = data['name']
        self.level = data['level']
        self.max_level = data['maxLevel']
        self.village = data['village']

    @property
    def is_max(self):
        """:class:`bool`: Helper property to tell you if the spell is the max level"""
        return self.max_level == self.level

    @property
    def is_builder_base(self):
        """:class:`bool`: Helper property to tell you if the spell belongs to the builder base"""
        return self.village == 'builderBase'

    @property
    def is_home_base(self):
        """:class:`bool`: Helper property to tell you if the spell belongs to the home base"""
        return self.village == 'home'


class WarAttack:
    """
    Represents a Clash of Clans War Attack

    Attributes
    -----------
    war:
        :class:`War` - The war this attack belongs to
    stars:
        :class:`int` - The stars achieved
    destruction:
        :class:`float` - The destruction achieved as a percentage (of 100)
    order:
        :class:`int` - The attack order in this war
    attacker_tag:
        :class:`int` - The attacker tag
    defender_tag:
        :class:`int` - The defender tag
    """
    __slots__ = ('war', 'member', 'stars',
                 'destruction', 'order',
                 'attacker_tag', 'defender_tag', '_data')

    def __init__(self, *, data, war, member):
        self._data = data

        self.war = war
        self.member = member
        self.stars = data['stars']
        self.destruction = data['destructionPercentage']
        self.order = data['order']
        self.attacker_tag = data['attackerTag']
        self.defender_tag = data['defenderTag']

    @property
    def attacker(self):
        """:class:`WarMember`: The attacker."""
        return self.war.get_member(tag=self.attacker_tag)

    @property
    def defender(self):
        """:class:`WarMember`: The defender."""
        return self.war.get_member(tag=self.defender_tag)


class Location:
    """Represents a Clash of Clans Location

    Attributes
    -----------
    id:
        :class:`str` - The location ID
    name:
        :class:`str` - The location name
    is_country:
        :class:`bool` - Indicates whether the location is a country
    country_code:
        :class:`str` - The shorthand country code, if the location is a country
    """
    __slots__ = ('id', 'name', 'is_country', 'country_code', '_data')

    def __init__(self, *, data):
        self._data = data

        self.id = data.get('id')
        self.name = data.get('name')
        self.is_country = data.get('isCountry')
        self.country_code = data.get('countryCode')

    def __str__(self):
        return self.name


class League:
    """Represents a Clash of Clans League

    Attributes
    -----------
    id:
        :class:`str` - The league ID
    name:
        :class:`str` - The league name
    """
    __slots__ = ('id', 'name', '_data')

    def __init__(self, *, data):
        self._data = data

        self.id = data.get('id')
        self.name = data.get('name')

    @property
    def badge(self):
        """:class:`Badge`: The league's badge"""
        return try_enum(Badge, data=self._data.get('iconUrls'))

    def __str__(self):
        return self.name


class LeagueRankedPlayer(BasicPlayer):
    """Represents a Clash of Clans League Ranked Player.
    Note that league season information is available only for Legend League.

    This class inherits both :class:`Player` and :class:`BasicPlayer`,
    and thus all attributes of these classes can be expected to be present.


    Attributes
    -----------
    rank:
        :class:`int` - The players rank in their league for this season
    """
    def __init__(self, *, data):
        self.rank = data.get('rank')
        super(LeagueRankedPlayer, self).__init__(data=data)


class Season:
    """Represents a Clash of Clans Player's Season.

    rank: """
    __slots__ = ('rank', 'trophies', 'id')

    def __init__(self, *, data):
        self.rank = data.get('rank')
        self.trophies = data.get('trophies')
        self.id = data.get('id')


class LegendStatistics:
    """Represents the Legend Statistics for a player.

    Attributes
    -----------
    player:
        :class:`Player` - The player
    legend_trophies:
        :class:`int` - The player's legend trophies
    """
    __slots__ = ('player', 'legend_trophies', '_data')

    def __init__(self, *, data, player):
        self._data = data

        self.player = player
        self.legend_trophies = data['legendTrophies']

    @property
    def current_season(self):
        """:class:`int`: Legend trophies for this season."""
        return try_enum(Season, data=self._data.get('currentSeason'))

    @property
    def previous_season(self):
        """:class:`int`: Legend trophies for the previous season."""
        return try_enum(Season, data=self._data.get('previousSeason'))

    @property
    def best_season(self):
        """:class:`int`: Legend trophies for the player's best season."""
        return try_enum(Season, data=self._data.get('bestSeason'))


class Badge:
    """Represents a Clash Of Clans Badge.

    Attributes
    -----------
    small:
        :class:`str` - URL for a small sized badge
    medium:
        :class:`str` - URL for a medium sized badge
    large:
        :class:`str` - URL for a large sized badge
    url:
        :class:`str` - Medium, the default URL badge size
    """
    __slots__ = ('small', 'medium', 'large', 'url', '_data')

    def __init__(self, *, data):
        # self._http = http
        self._data = data

        self.small = data.get('small')
        self.medium = data.get('medium')
        self.large = data.get('large')

        self.url = self.medium

    async def save(self, fp, size=None):
        """This funtion is a coroutine. Save this badge as a file-like object.

        :param fp: :class:`os.PathLike`
                    The filename to save the badge to
        :param size: Optional[:class:`str`] Either `small`, `medium` or `large`. The default is `medium`

        :raise HTTPException: Saving the badge failed

        :raise NotFound: The url was not found

        :return: :class:`int` The number of bytes written
        """
        sizes = {'small': self.small,
                 'medium': self.medium,
                 'large': self.large}

        if size and size in sizes.keys():
            url = sizes[size]
        else:
            url = self.medium

        # data = self._http.get_data_from_url(url)
        #
        # with open(fp, 'wb') as f:
        #     return f.write(data)


class Timestamp:
    """Represents a Clash of Clans Timestamp

    Attributes
    -----------
    time: :class:`str`: The raw timestamp string (ISO8601) as given by the API.
    """
    __slots__ = ('time', '_data')

    def __init__(self, *, data):
        self._data = data

        self.time = data

    @property
    def utc_timestamp(self):
        """:class:`datetime`: The timestamp as a UTC datetime object"""
        return from_timestamp(self.time)

    @property
    def now(self):
        """:class:`datetime`: The time in UTC now as a datetime object"""
        return datetime.utcnow()

    @property
    def seconds_until(self):
        """:class:`int`: Number of seconds until the timestamp. This may be negative."""
        delta = self.utc_timestamp - self.now
        return delta.total_seconds()


class LeaguePlayer:
    """Represents a Clash of Clans League Player

    Attributes
    -----------
    tag:
        :class:`str` - The player's tag
    name:
        :class:`str` - The player's name
    town_hall:
        :class:`int` - The player's town hall level"""

    __slots__ = ('tag', 'name', 'town_hall', '_data')

    def __str__(self):
        return self.name

    def __init__(self, *, data):
        self._data = data

        self.tag = data.get('tag')
        self.name = data.get('name')
        self.town_hall = data.get('townHallLevel')


class LeagueClan(BasicClan):
    """Represents a Clash of Clans League Clan

    This class inherits both :class:`Clan` and :class:`BasicClan`,
    and thus all attributes of these classes can be expected to be present.

    Attributes
    -----------

    """
    def __init__(self, *, data):
        super(LeagueClan, self).__init__(data=data)

    @property
    def _members(self):
        """|iter|

        Returns an iterable of :class:`LeaguePlayer`: all players participating in this league season"""
        return iter(LeaguePlayer(data=mdata) for mdata in self._data.get('members', []))

    @property
    def members(self):
        """List[:class:`LeaguePlayer`} A list of players participating in this league season"""
        return list(self._members)


class LeagueGroup:
    """Represents a Clash of Clans League Group

    Attributes
    -----------
    state:
        :class:`str` - The current state of the league group (`inWar`, `preparation` etc.)
    season:
        :class:`str` - The current season of the league group
    """
    __slots__ = ('state', 'season', '_rounds', '_data')

    def __init__(self, *, data):
        self._data = data

        self._rounds = []
        self.state = data.get('state')
        self.season = data.get('season')

    @property
    def _clans(self):
        """|iter|

        Returns an iterable of class:`LeagueClan`: all participating clans"""
        return iter(LeagueClan(data=cdata) for cdata in self._data.get('clans', []))

    @property
    def clans(self):
        """List[class:`LeagueClan`]: A list of participating clans."""
        return list(self._clans)

    @property
    def rounds(self):
        """List[List[]]: A list of lists containing all war tags for each round"""
        # TODO: Find a better method of doing this
        rounds = []
        for rdata in self._data.get('rounds', []):
            rounds.append(rdata['warTags'])

        return rounds


class LeagueWar(CurrentWar):
    """Represents a Clash of Clans LeagueWar

    This class inherits both :class:`BaseWar` and :class:`CurrentWar`,
    and thus all attributes of these classes can be expected to be present.

    Attributes
    -----------
    tag:
        :class:`str` - The war tag
    """
    def __init__(self, *, data):
        self.tag = data.get('tag')
        super(LeagueWar, self).__init__(data=data)


class LeagueWarLogEntry:
    """Represents a Clash of Clans War Log entry for a League Season

    Attributes
    -----------
    end_time:
        :class:`Timestamp` - The end time of the war as a Timestamp object
    team_size:
        :class:`int` - The number of players per clan in war
    clan:
        :class:`Clan` - The offensive clan. Note this is only a :class:`Clan`, unlike that of a :class:`WarLog`
    enemy_stars:
        :class:`int` - Total enemy stars for all wars
    attack_count:
        :class:`int` - The total attacks completed by your clan over all wars
    stars:
        :class:`int` The total stars by your clan over all wars
    destruction:
        :class:`float` - The total destruction by your clan over all wars
    clan_level:
        :class:`int` - Your clan level.
    """

    __slots__ = ('end_time', 'team_size', 'clan', 'enemy_stars',
                 'attack_count', 'stars', 'destruction', 'clan_level')

    def __init__(self, *, data):
        self.end_time = try_enum(Timestamp, data.get('endTime'))
        self.team_size = data.get('teamSize')
        self.clan = try_enum(Clan, data.get('clan'))
        try:
            self.enemy_stars = data['opponent']['stars']
        except KeyError:
            self.enemy_stars = None

        if self.clan:
            self.attack_count = self.clan._data.get('attacks')
            self.stars = self.clan._data.get('stars')
            self.destruction = self.clan._data.get('destructionPercentage')
            self.clan_level = self.clan._data.get('clanLevel')



# -*- coding: utf-8 -*-

__version__ = '0.1.1'

from .client import Client
from .dataclasses import (
    Clan,
    BasicClan,
    SearchClan,
    WarClan,
    Player,
    BasicPlayer,
    WarMember,
    SearchPlayer,
    BaseWar,
    WarLog,
    CurrentWar,
    Achievement,
    Troop,
    Hero,
    Spell,
    WarAttack,
    Location,
    League,
    LeagueRankedPlayer,
    Season,
    LegendStatistics,
    Badge,
    Timestamp,
    LeaguePlayer,
    LeagueClan,
    LeagueGroup,
    LeagueWar,
    LeagueWarLogEntry
)
from .errors import (
    ClashOfClansException,
    HTTPException,
    NotFound,
    InvalidArgument,
    InvalidCredentials,
    Forbidden,
    Maitenance
)
from .http import HTTPClient

from .enums import (
    CacheType
)



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 mathsman5133

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""


class ClashOfClansException(Exception):
    """Base exception for coc.py
    """


class HTTPException(ClashOfClansException):
    """Base exception for when a HTTP request fails

    Attributes
    -----------
    response:
        :class:`aiohttp.ClientResponse` - The response of the failed HTTP request.

    status:
        :class:`int` - The status code of the HTTP request

    reason:
        :class:`str` - The reason provided by the API.
            This could be an empty string if nothing was given.

    message:
        :class:`str` - The more detailed message provided by the API.
            This could be an empty string if nothing was given

    """
    def __init__(self, response, message):
        self.response = response
        self.status = response.status
        try:
            self.reason = message.get('reason', 'Unknown')
            self.message = message.get('message', '')
        except Exception:
            self.reason = 'Unknown'
            self.message = ''

        fmt = '{0.reason} (status code: {0.status})'
        if len(self.message):
            fmt = fmt + ' :{1}'

        super().__init__(fmt.format(self.response, self.message))


class InvalidArgument(ClashOfClansException):
    """
    Thrown when an error status 400 occurs.

    Client provided incorrect parameters for the request.

    Subclass of :exc:`HTTPException`
    """
    pass


class InvalidCredentials(HTTPException):
    """Thrown when an error status 403 occurs and the reason is invalid credentials.

    Special Exception thrown when missing/incorrect credentials
    were passed. This is when your email/password pair is incorrect.
    Subclass of :exc:`HTTPException`
    """

    pass


class Forbidden(HTTPException):
    """Thrown when an error status 403 occurs.

    API token does not grant access to the requested resource.

    Subclass of :exc:`HTTPException`"""


class NotFound(HTTPException):
    """Thrown when an error status 404 occurs.

    The resource was not found.

    Subclass of :exc:`HTTPException`
    """
    pass


class Maitenance(HTTPException):
    """Thrown when an error status 503 occurs.

    Service is temporarily unavailable because of maintenance.

    Subclass of :exc:`HTTPException`
    """
    pass



from enum import Enum


class CacheType(Enum):
    search_clans = 'cache_search_clans'
    war_clans = 'cache_war_clans'

    cache_search_players = 'cache_search_players'
    war_players = 'cache_war_players'

    current_wars = 'cache_current_wars'
    war_logs = 'cache_war_logs'

    league_groups = 'cache_league_groups'
    league_wars = 'cache_league_wars'

    locations = 'cache_locations'
    leagues = 'cache_leagues'
    seasons = 'cache_seasons'

    def __str__(self):
        return self.value



import re
import json

from datetime import datetime


def to_json(model):
    dct = {}
    for attr in dir(model):
        # ignore private methods
        if attr.startswith('_'):
            continue

        value = getattr(model, attr)

        # iterate through lists - may be a list of objects eg. members in war
        if isinstance(value, list):
            for i in value:
                if not isinstance(i, (list, str, bool)):
                    dct[attr] = [to_json(i)] if attr not in dct.keys() else dct[attr].append(to_json(i))
                    continue

                dct[attr] = [to_json(i)] if attr not in dct.keys() else dct[attr].append(to_json(i))

            continue

        # if it is an object attribute
        if not isinstance(value, (list, str, bool)):
            dct[attr] = [to_json(value)] if attr not in dct.keys() else dct[attr].append(to_json(value))
            continue

        # just a bool, string or int now so can safely add to dict
        dct[attr] = [to_json(value)] if attr not in dct.keys() else dct[attr].append(to_json(value))

    return json.loads(dct, separators=(',', ':'), ensure_ascii=True)


def find(predicate, seq):
    for element in seq:
        if predicate(element):
            return element
    return None


def get(iterable, **attrs):
    def predicate(elem):
        for attr, val in attrs.items():
            nested = attr.split('__')
            obj = elem
            for attribute in nested:
                obj = getattr(obj, attribute)

            if obj != val:
                return False
        return True

    return find(predicate, iterable)


def from_timestamp(timestamp):
    return datetime.strptime(timestamp, '%Y%m%dT%H%M%S.000Z')


def correct_tag(tag, prefix='#'):
    """Attempts to correct malformed Clash of Clans tags
    to match how they are formatted in game
    
    Example
    ---------
        ' 123aBc O' -> '#123ABC0'
    """
    return prefix + re.sub(r'[^A-Z0-9]+', '', tag.upper()).replace('O', '0')



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 mathsman5133

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import asyncio
import logging


from .cache import Cache
from .http import HTTPClient
from .dataclasses import *

log = logging.getLogger(__name__)

LEAGUE_WAR_STATE = 'notInWar'
KEY_MINIMUM, KEY_MAXIMUM = 1, 10


class Client:
    """
    This is the client connection used to interact with the Clash of Clans API.

    .. _event loop: https://docs.python.org/3/library/asyncio-eventloops.html

    Parameters
    -------------

    email: :class:`str`
        Your password email from https://developer.clashofclans.com
        This is used when updating the key automatically if your IP changes

    password: :class:`str`
        Your password login from https://developer.clashofclans.com
        This is used when updating the key automatically if your IP changes

    key_count: Optional[:class:`int`]
        The amount of keys to use for this client. Maximum of 10.
        Defaults to 1

    key_names: Optional[:class:`str`]
        Default name for keys created to use for this client.
        All keys created or to be used with this client must
        have this name.

        Defaults to "Created with coc.py Client"

    throttle_limit: Optional[:class:`int`]
        The number of requests per token per second to send to the API.
        Once hitting this limit, the library will automatically throttle
        your requests.

        .. note::

            Setting this value too high may result in the API rate-limiting
            your requests. This means you cannot request for ~30-60 seconds.

        .. warning::

            Setting this value too high may result in your requests being
            deemed "API Abuse", potentially resulting in an IP ban.


        Defaults to 10 requests per token, per second.

    loop: Optional[event loop]
        The `event loop` to use for HTTP requests.
        An ``asyncio.get_event_loop()`` will be used if none is passed

    Attributes
    -----------
    loop
        The `event_loop`_ that is used for HTTP requests

    """
    cache_search_clans = Cache()
    cache_war_clans = Cache()

    cache_search_players = Cache()
    cache_war_players = Cache()

    cache_current_wars = Cache()
    cache_war_logs = Cache()

    cache_league_groups = Cache()
    cache_league_wars = Cache()

    cache_locations = Cache()

    cache_leagues = Cache()
    cache_seasons = Cache()

    def __init__(self, email, password, key_count=1,
                 key_names='Created with coc.py Client',
                 throttle_limit=10,
                 loop=None):

        self.loop = loop or asyncio.get_event_loop()
        correct_key_count = max(min(KEY_MAXIMUM, key_count), KEY_MINIMUM)

        if not key_count == correct_key_count:
            raise RuntimeError("Key count must be within {}-{}".format(
                KEY_MINIMUM, KEY_MAXIMUM))

        self.http = HTTPClient(client=self, email=email, password=password,
                               key_names=key_names, loop=self.loop,
                               key_count=correct_key_count,
                               throttle_limit=throttle_limit)

        log.info('Clash of Clans API client created')

    async def close(self):
        """
        Closes the HTTP connection
        """
        log.info('Clash of Clans client logging out...')
        await self.http.close()

    async def on_key_reset(self, new_key):
        """Event: called when one of the client's keys are reset.

        By default this does nothing.

        Example
        ---------

        You can manually override this by either:

        Subclassing Client:

        .. code-block:: python3

            class Client(coc.Client):
                def __init__(self, key, email, password):
                    super().__init__(email, password, key_count=1,
                                 key_names='Created with coc.py Client', loop=None)

                def on_key_reset(key):
                    print('My new key is {}'.format(key))

        Using the event decorator:

        .. code-block:: python3

            @client.event()
            async def on_key_reset(key):
                print('My new key is {}'.format(key))

        :param new_key: :class:`str` The new key
        """
        pass

    def event(self, fctn):
        """A decorator that registers an event.

        The only event at present is :func:`on_key_reset`.

        This could be a coro or regular function.

        Example
        --------

        .. code-block:: python3

            @client.event()
            async def on_key_reset(key):
                print('My new key is {}'.format(key))
        """
        setattr(self, fctn.__name__, fctn)
        log.info('Successfully registered %s event', fctn.__name__)
        return fctn

    def dispatch(self, event_name, *args, **kwargs):
        log.debug('Dispatching %s event', event_name)
        event = 'on_' + event_name

        try:
            fctn = getattr(self, event)
        except AttributeError:
            return
        else:
            if asyncio.iscoroutinefunction(fctn):
                asyncio.ensure_future(fctn(*args, **kwargs), loop=self.loop)
            else:
                event(*args, **kwargs)

    def set_cache(self, *cache_to_edit, max_size=128, expiry=None):
        """Set the max size and expiry time for a cached object.

        .. note::

            Calling this method will override and create a new cache instance,
            removing all previously cached objects


        :param cache_to_edit: :class:`str` or :class:`CacheType` enum. The name of cache type to change.
        :param max_size: :class:`int` The max size of the created cache. Defaults to 128
        :param expiry: :class:`int` The expiry time in seconds of the cache. Defaults to None (cache does not expire)
        """
        for cache_type in cache_to_edit:
            if not getattr(self, cache_type):
                raise ValueError('{} is not a valid cached data class type'.format(cache_to_edit))

            cache = Cache(max_size=max_size, ttl=expiry)
            log.debug('Cache type %s has been set with max size %s and expiry %s seconds',
                      cache_type, max_size, expiry)
            setattr(Client, cache_type, cache)

    async def search_clans(self, *, name: str=None, war_frequency: str=None,
                           location_id: int = None, min_members: int=None, max_members: int=None,
                           min_clan_points: int = None, min_clan_level: int=None, limit: int=None,
                           before: int=None, after: int=None
                           ):
        """Search all clans by name and/or filtering the results using various criteria.

        At least one filtering criteria must be defined and if name is used as part
        of search, it is required to be at least three characters long.

        :param name: Optional[:class:`str`]
        :param war_frequency: Optional[:class:`str`]
        :param location_id: Optional[:class:`int`]
        :param min_members: Optional[:class:`int`]
        :param max_members: Optional[:class:`int`]
        :param min_clan_points: Optional[:class:`int`]
        :param min_clan_level: Optional[:class:`int`]
        :param limit: Optional[:class:`int`]

        :raise HTTPException: no options were passed

        :return: :class:`list` of all :class:`SearchClan` found

        """

        r = await self.http.search_clans(name=name, warFrequency=war_frequency, locationId=location_id,
                                         minMembers=min_members, maxMembers=max_members, minClanPoints=min_clan_points,
                                         minClanLevel=min_clan_level, limit=limit, before=before, after=after)

        clans = list(SearchClan(data=n, http=self.http) for n in r.get('items', []))

        return clans

    @cache_search_clans.get_cache()
    async def get_clan(self, tag, cache=False, fetch=True):
        """Get information about a single clan by clan tag. Clan tags can be found using clan search operation.

        :param tag: :class:`str` The tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`SearchClan`
        """
        r = await self.http.get_clan(tag)
        return SearchClan(data=r, http=self.http)

    async def get_clans(self, tags, cache=False, fetch=True):
        """Get information about multiple clans by clan tag. Refer to `Client.get_clan` for more information.
        
        Example
        ---------

        .. code-block:: python3
        
            tags = [...]
            async for clan in Client.get_clans(tags):
                print(clan)
                
        :param tags: :class:`collections.Iterable` The tags to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`SearchClan`
        """
        
        tags = iter(tags)
        
        for tag in tags:
            if cache:
                c = self.cache_search_clans.get(tag)
                if c:
                    yield c
                if fetch is False:
                    yield None

            r = await self.http.get_clan(tag)
            c = SearchClan(data=r, http=self.http)
            self.cache_search_clans.add(c.tag, c)

            yield c

    async def get_members(self, clan_tag, cache=False, fetch=True):
        """List clan members. This is equivilant to `Client.get_clan('tag').members.

        :param clan_tag: :class:`str` The clan tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`list` of :class:`BasicPlayer`
        """
        if cache:
            c = self.cache_search_clans.get(clan_tag)
            if c:
                return c.members

            if fetch is False:
                return None

        r = await self.http.get_clan(clan_tag)
        clan = SearchClan(data=r, http=self.http)

        self.cache_search_clans.add(clan.tag, clan)

        return clan.members

    @cache_war_logs.get_cache()
    async def get_warlog(self, clan_tag, cache=False, fetch=True):
        """Retrieve clan's clan war log

        :param clan_tag: :class:`str` The tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`list` of either :class:`WarLog` or :class:`LeagueWarLogEntry`, according to which war it is,

        """
        r = await self.http.get_clan_warlog(clan_tag)

        wars = []
        for n in r.get('items', []):
            # lately war log entries for sccwl can be distinguished by a `null` result
            if n.get('result') is None:
                wars.append(LeagueWarLogEntry(data=n))
                continue

            # for earlier logs this is distinguished by no opponent tag (result called `tie`)
            if n.get('opponent', {}).get('tag', None) is None:
                wars.append(LeagueWarLogEntry(data=n))
                continue

            wars.append(WarLog(data=n))

        self.cache_war_logs.add(wars[0].clan.tag, wars)

        return wars

    @cache_current_wars.get_cache()
    async def get_current_war(self, clan_tag, cache=False, fetch=True):
        """
        Retrieve information about clan's current clan war

        :param clan_tag: :class:`str` The clan tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`CurrentWar`
        """
        r = await self.http.get_clan_current_war(clan_tag)
        return CurrentWar(data=r)
    
    async def get_current_wars(self, clan_tags, cache=False, fetch=True):
        """
        Retrieve information multiple clan's current clan wars
        
        Example
        ---------

        .. code-block:: python3
        
            tags = [...]
            async for clan_war in Client.get_current_wars(tags):
                print(clan_war.opponent)

        :param clan_tags: :class:`collections.Iterable` The clan tags to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`CurrentWar`
        """
        
        clan_tags = iter(clan_tags)
        
        for clan_tag in clan_tags:
            if cache:
                wl = self.cache_current_wars.get(clan_tag)
                if wl:
                    yield wl

                if fetch is False:
                    yield None

            r = await self.http.get_clan_current_war(clan_tag)
            war = CurrentWar(data=r)

            self.cache_current_wars.add(war.clan.tag, war)
            yield war

    @cache_league_groups.get_cache()
    async def get_league_group(self, clan_tag, cache=False, fetch=True):
        """Retrieve information about clan's current clan war league group

        :param clan_tag: :class:`str` The tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`LeagueGroup`
        """
        r = await self.http.get_clan_war_league_group(clan_tag)
        return LeagueGroup(data=r)

    @cache_league_wars.get_cache()
    async def get_league_war(self, war_tag, cache=False, fetch=True):
        """
        Retrieve information about a clan war league war

        :param war_tag: :class:`str` The tag of league war to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`LeagueWar`
        """
        r = await self.http.get_cwl_wars(war_tag)
        return LeagueWar(data=r)

    # locations
    async def search_locations(self, *, limit: int=10,
                               before: int=None, after: int=None):
        """
        List all available locations

        :param limit: Optional[:class:`int`] Number of items to fetch. Default is 10

        :raise HTTPException: no options were passed

        :return: :class:`list` of all :class:`Location` found

        """

        data = await self.http.search_locations(limit=limit, before=before, after=after)

        locations = list(Location(data=n) for n in data['items'])

        for n in locations:
            self.cache_locations.add(n.id, n)

        return locations

    @cache_locations.get_cache()
    async def get_location(self, location_id, cache=False, fetch=True):
        """
        Get information about specific location

        :param location_id: :class:`int` The location ID to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`Location`
        """
        r = await self.http.get_location(location_id)
        return Location(data=r)

    async def get_location_clan(self, location_id, *, limit: int=None,
                                before: int=None, after: int=None):
        """
        Get clan rankings for a specific location


        :param location_id: :class:`int` The location ID to search
        :param limit: :class:`int` The number of results to fetch

        :return: :class:`Clan`
        """

        r = await self.http.get_location_clans(location_id, limit=limit, before=before, after=after)
        return list(Clan(data=n) for n in r['items'])

    async def get_location_players(self, location_id, *, limit: int=None,
                                   before: int=None, after: int=None):
        """
        Get player rankings for a specific location

        :param location_id: :class:`int` The location ID to search
        :param limit: :class:`int` The number of results to fetch

        :return: :class:`Player`
        """
        r = await self.http.get_location_players(location_id, limit=limit, before=before, after=after)
        return list(Player(data=n) for n in r['items'])

    async def get_location_clans_versus(self, location_id, *, limit: int=None,
                                        before: int=None, after: int=None):
        """
        Get clan versus rankings for a specific location

        :param location_id: :class:`int` The location ID to search
        :param limit: :class:`int` The number of results to fetch

        :return: :class:`Clan`
        """
        r = await self.http.get_location_clans_versus(location_id, limit=limit, before=before, after=after)
        return list(Clan(data=n) for n in r['items'])

    async def get_location_players_versus(self, location_id, *, limit: int = None,
                                          before: int = None, after: int = None):
        """
        Get player versus rankings for a specific location

        :param location_id: :class:`int` The location ID to search
        :param limit: :class:`int` The number of results to fetch

        :return: :class:`Player`
        """
        r = await self.http.get_location_players_versus(location_id, limit=limit, before=before, after=after)
        return list(Player(data=n) for n in r['items'])

    # leagues

    async def search_leagues(self, *, limit: int=None, before: int=None, after: int=None):
        """
        Get list of leagues.

        :param limit: Optional[:class:`int`] Number of items to fetch. Default is 10

        :raise HTTPException: no options were passed

        :return: :class:`list` of all :class:`League` found

        """
        r = await self.http.search_leagues(limit=limit, before=before, after=after)
        leagues = list(League(data=n) for n in r['items'])

        for n in leagues:
            self.cache_leagues.add(n.id, n)

        return leagues

    @cache_leagues.get_cache()
    async def get_league(self, league_id, cache=False, fetch=True):
        """
        Get league information

        :param league_id: :class:`str` The ID to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`League`
        """
        r = await self.http.get_league(league_id)
        return League(data=r)

    async def get_seasons(self, league_id):
        """Get league seasons. Note that league season information is available only for Legend League.

        :param league_id: :class:`str` The league ID to search for

        :return: :class:`dict`

        In the form

        .. code-block:: json

            {
                "id": "string"
            }

        where `id` is the season ID
        """
        r = await self.http.get_league_seasons(league_id)
        return r['items']

    async def get_season_rankings(self, league_id, season_id, cache=False, fetch=True):
        """Get league season rankings.
        Note that league season information is available only for Legend League.

        :param league_id: :class:`str` The league ID to search for
        :param season_id: :class:`str` The season ID to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`list` of :class:`LeagueRankedPlayer`
        """
        if cache:
            try:
                data = self.cache_seasons.get(league_id)[season_id]
                if data:
                    return data

            except KeyError:
                pass

            if fetch is False:
                return None

        r = await self.http.get_league_season_info(league_id, season_id)
        players = list(LeagueRankedPlayer(data=n) for n in r.get('items', []))

        self.cache_seasons.add(league_id, {season_id: players})

        return players

    # players
    @cache_search_players.get_cache()
    async def get_player(self, player_tag, cache=False, fetch=True):
        """Get information about a single player by player tag.
        Player tags can be found either in game or by from clan member lists.

        :param player_tag: :class:`str` The player tag to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`SearchPlayer`
        """
        r = await self.http.get_player(player_tag)
        return SearchPlayer(data=r)

    async def get_players(self, player_tags, cache=False, fetch=True):
        """Get information about a multiple players by player tag.
        Player tags can be found either in game or by from clan member lists.
        
        Example
        ---------

        .. code-block:: python3
        
            tags = [...]
            async for player in Client.get_players(tags):
                print(player)

        :param player_tags: :class:`collections.Iterable` The player tags to search for
        :param cache: Optional[:class:`bool`] Indicates whether to search the cache before making an HTTP request
        :param fetch: Optional[:class:`bool`] Indicates whether an HTTP call should be made if cache is empty.
                        Defaults to ``True``. If this is ``False`` and item in cache was not found,
                        ``None`` will be returned

        :return: :class:`SearchPlayer`
        """
        
        player_tags = iter(player_tags)
        
        for player_tag in player_tags:
            if cache:
                data = self.cache_search_players.get(player_tag)
                if data:
                    yield data

                if fetch is False:
                    yield None

            r = await self.http.get_player(player_tag)
            p = SearchPlayer(data=r)
            self.cache_search_players.add(p.tag, p)

            yield p




from setuptools import setup
import os

requirements = []
with open(os.path.join(os.getcwd(), 'requirements.txt')) as f:
    requirements = f.read().splitlines()

version = '0.1.1'

readme = ''
with open('README.rst') as f:
    readme = f.read()

setup(name='coc.py',
      author='mathsman5133',
      url='https://github.com/mathsman5133/coc.py',
      packages=['coc'],
      version=version,
      license='MIT',
      description='A python wrapper for the Clash of Clans API',
      long_description=readme,
      long_description_content_type="text/x-rst",
      python_requires='>=3.5.3',
      install_requires=requirements,
      extra_requires={'docs': ['sphinx==1.7.4', 'sphinx_rtd_theme']}
      )



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 mathsman5133

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""


import json
import logging
import aiohttp
import asyncio

from urllib.parse import urlencode
from itertools import cycle
from datetime import datetime
from collections import deque

import coc.nest_asyncio
from .errors import HTTPException, Maitenance, NotFound, InvalidArgument, Forbidden, InvalidCredentials

log = logging.getLogger(__name__)
KEY_MINIMUM, KEY_MAXIMUM = 1, 10


async def json_or_text(response):
    try:
        ret = await response.json()
    except aiohttp.ContentTypeError:
        ret = await response.text(encoding='utf-8')

    return ret


class Throttler:
    """Simple throttler for asyncio"""

    def __init__(self, rate_limit, retry_interval=0.001, loop=None):
        self.rate_limit = rate_limit
        self.retry_interval = retry_interval
        self.loop = loop or asyncio.get_event_loop()

        self._task_logs = deque()

    async def __aenter__(self):
        while True:
            now = self.loop.time()

            # Pop items(which are start times) that are no longer in the
            # time window
            while self._task_logs:
                if now - self._task_logs[0] > 1.0:
                    self._task_logs.popleft()
                else:
                    break

            # Exit the infinite loop when new task can be processed
            if len(self._task_logs) < self.rate_limit:
                break

            log.info('Request throttled. Sleeping for %s seconds.', self.retry_interval)
            await asyncio.sleep(self.retry_interval)

        # Push new task's start time
        self._task_logs.append(self.loop.time())

        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass


class Route:
    BASE = 'https://api.clashofclans.com/v1'
    API_PAGE_BASE = 'https://developer.clashofclans.com/api'

    def __init__(self, method, path, kwargs: dict=None, api_page=False):
        if '#' in path:
            path = path.replace('#', '%23')

        self.method = method
        self.path = path
        url = (self.API_PAGE_BASE + self.path if api_page else self.BASE + self.path)

        if kwargs:
            self.url = '{}?{}'.format(url, urlencode({k: v for k, v in kwargs.items() if v is not None}))
        else:
            self.url = url


class HTTPClient:
    def __init__(self, client, loop, email, password,
                 key_names, key_count, throttle_limit):
        self.client = client
        coc.nest_asyncio.apply(loop)
        self.loop = loop
        self.email = email
        self.password = password
        self.key_names = key_names
        self.key_count = key_count
        self.throttle_limit = throttle_limit

        per_second = key_count * throttle_limit

        self.__lock = asyncio.Semaphore(per_second)
        self.__throttle = Throttler(per_second, loop=self.loop)

        loop.run_until_complete(self.get_keys())

    async def get_keys(self):
        self.__session = aiohttp.ClientSession(loop=self.loop)

        key_count = self.key_count
        ip = await self.get_ip()
        response_dict, session = await self.login_to_site(self.email, self.password)
        cookies = self.create_cookies(response_dict, session)
        current_keys = (await self.find_site_keys(cookies))['keys']
        self._keys = [key['key'] for key in current_keys if key['name'] == self.key_names]
        available_keys = KEY_MAXIMUM - len(current_keys)

        if len(self._keys) <= key_count:
            keys_needed = key_count - len(self._keys)
            if available_keys >= keys_needed:
                for _ in range(keys_needed):
                    key_description = "Created on {}".format(datetime.now().strftime('%c'))
                    self._keys.append(await self.create_key(
                        cookies, self.key_names, key_description, [ip]))
            else:
                await self.close()
                raise RuntimeError(("There are {} API keys already created and {} match \"{}\" out of a max of "
                                    "{} please goto {} and delete unused keys or rename to \"{}\".").format(
                    len(current_keys), len(self._keys),
                    self.key_names, KEY_MAXIMUM,
                    'https://developer.clashofclans.com',
                    self.key_names))

        self.keys = cycle(self._keys)

    async def close(self):
        if self.__session:
            await self.__session.close()

    async def request(self, route, **kwargs):
        method = route.method
        url = route.url

        if 'headers' not in kwargs:
            key = next(self.keys)

            headers = {
                    "Accept": "application/json",
                    "authorization": "Bearer {}".format(key),
                }
            kwargs['headers'] = headers

        if 'json' in kwargs:
            kwargs['headers']['Content-Type'] = 'application/json'

        async with self.__lock:
            async with self.__throttle, self.__session.request(method, url, **kwargs) as r:
                log.debug('%s (%s) has returned %s', url, method, r.status)
                data = await json_or_text(r)
                log.debug(data)

                if 200 <= r.status < 300:
                    log.debug('%s has received %s', url, data)
                    return data

                if r.status == 400:
                    raise InvalidArgument(r, data)

                if r.status == 403:
                    if data.get('reason') in ['accessDenied.invalidIp']:
                        if 'key' in locals():
                            await self.reset_key(key)
                            log.info('Reset Clash of Clans key')
                            return await self.request(route, **kwargs)

                    raise Forbidden(r, data)

                if r.status == 404:
                    raise NotFound(r, data)
                if r.status == 429:
                    log.error('We have been rate-limited by the API. '
                              'Reconsider the number of requests you are allowing per second.')
                    raise HTTPException(r, data)

                if r.status == 503:
                    raise Maitenance(r, data)
                else:
                    raise HTTPException(r, data)

    async def get_ip(self):
        async with self.__session.request('GET', 'http://ip.42.pl/short') as r:
            log.debug('%s (%s) has returned %s', 'http://ip.42.pl/short', 'GET', r.status)
            ip = await r.text()
            log.debug('%s has received %s', 'http://ip.42.pl/short', ip)
        return ip

    @staticmethod
    def create_cookies(response_dict, session):
        return "session={};game-api-url={};game-api-token={}".format(
            session,
            response_dict['swaggerUrl'],
            response_dict['temporaryAPIToken']
        )

    async def reset_key(self, key):
        ip = await self.get_ip()
        # TODO: Distinguish/recognise by key name rather than current ip for running clients on multiple IPs
        # should probably put something else in here
        # to distinguish each key like a date
        key_name = 'coc.py created key'
        # Also, probably fix this as well
        key_description = 'learn more about this project at: https://github.com/mathsman5133/coc.py'
        whitelisted_ips = [ip]

        response_dict, session = await self.login_to_site(self.email, self.password)
        cookies = self.create_cookies(response_dict, session)

        existing_keys = (await self.find_site_keys(cookies))['keys']
        key_id = [t['id'] for t in existing_keys if t['key'] == key]

        await self.delete_key(cookies, key_id)

        new_key = await self.create_key(cookies, key_name, key_description, whitelisted_ips)

        # this is to prevent reusing an already used keys.
        # All it does is move the current key to the front,
        # by moving any already used ones to the end so
        # we keep the original key order moving forward.
        keys = self._keys
        key_index = keys.index(key)
        self._keys = keys[key_index:] + keys[:key_index]

        # now we can set the new key which is the first
        # one in self._keys, then start the cycle over.
        self._keys[0] = new_key
        self.keys = cycle(self.keys)
        self.client.dispatch('key_reset', new_key)

    # clans

    def search_clans(self, **kwargs):
        return self.request(Route('GET', '/clans', kwargs))

    def get_clan(self, tag):
        return self.request(Route('GET', '/clans/{}'.format(tag), {}))

    def get_clan_members(self, tag):
        return self.request(Route('GET', '/clans/{}/members'.format(tag), {}))

    def get_clan_warlog(self, tag):
        return self.request(Route('GET', '/clans/{}/warlog'.format(tag), {}))

    def get_clan_current_war(self, tag):
        return self.request(Route('GET', '/clans/{}/currentwar'.format(tag), {}))

    def get_clan_war_league_group(self, tag):
        return self.request(Route('GET', '/clans/{}/currentwar/leaguegroup'.format(tag), {}))

    def get_cwl_wars(self, war_tag):
        return self.request(Route('GET', '/clanwarleagues/wars/'.format(war_tag), {}))

    # locations

    def search_locations(self, **kwargs):
        return self.request(Route('GET', '/locations', kwargs))

    def get_location(self, location_id):
        return self.request(Route('GET', '/locations/{}'.format(location_id), {}))

    def get_location_clans(self, location_id, **kwargs):
        return self.request(Route('GET', '/locations/{}/rankings/clans'.format(location_id), kwargs))

    def get_location_players(self, location_id, **kwargs):
        return self.request(Route('GET', '/locations/{}/rankings/players'.format(location_id), kwargs))

    def get_location_clans_versus(self, location_id, **kwargs):
        return self.request(Route('GET', '/locations/{}/rankings/clans-versus'.format(location_id), kwargs))

    def get_location_players_versus(self, location_id, **kwargs):
        return self.request(Route('GET', '/locations/{}/rankings/players-versus'.format(location_id), kwargs))

    # leagues

    def search_leagues(self, **kwargs):
        return self.request(Route('GET', '/leagues', kwargs))

    def get_league(self, league_id):
        return self.request(Route('GET', '/leagues/{}'.format(league_id), {}))

    def get_league_seasons(self, league_id, **kwargs):
        return self.request(Route('GET', '/leagues/{}/seasons'.format(league_id), kwargs))

    def get_league_season_info(self, league_id, season_id, **kwargs):
        return self.request(Route('GET', '/leagues/{}/seasons/{}'.format(league_id, season_id), kwargs))

    # players

    def get_player(self, player_tag):
        return self.request(Route('GET', '/players/{}'.format(player_tag), {}))

    # key updating management

    async def login_to_site(self, email, password):
        login_data = {
            'email': email,
            'password': password
                      }
        headers = {
            'content-type': 'application/json'
        }
        async with self.__session.post('https://developer.clashofclans.com/api/login',
                                       json=login_data, headers=headers) as sess:
            response_dict = await sess.json()
            log.debug('%s has received %s', 'https://developer.clashofclans.com/api/login',
                      response_dict)
            if sess.status == 403:
                raise InvalidCredentials(sess, response_dict)

            session = sess.cookies.get('session').value

        return response_dict, session

    async def find_site_keys(self, cookies):
        headers = {
            "cookie": cookies,
            "content-type": "application/json"
        }
        async with self.__session.post('https://developer.clashofclans.com/api/apikey/list',
                                       data=json.dumps({}), headers=headers) as sess:
            existing_keys_dict = await sess.json()
            log.debug('%s has received %s', 'https://developer.clashofclans.com/api/apikey/list',
                      existing_keys_dict)

        return existing_keys_dict

    async def create_key(self, cookies, key_name, key_description, cidr_ranges):
        headers = {
            "cookie": cookies,
            "content-type": "application/json"
        }

        data = {
            "name": key_name,
            "description": key_description,
            "cidrRanges": cidr_ranges
        }

        r = await self.request(Route('POST', '/apikey/create', api_page=True), json=data, headers=headers)
        return r['key']['key']

    def delete_key(self, cookies, key_id):
        headers = {
            "cookie": cookies,
            "content-type": "application/json"
        }

        data = {
            "id": key_id
        }

        return self.request(Route('POST', '/apikey/revoke', api_page=True), json=data, headers=headers)

    async def get_data_from_url(self, url):
        async with self.__session.get(url) as r:
            if r.status == 200:
                return await r.read()
            elif r.status == 404:
                raise NotFound(r, 'image not found')
            else:
                raise HTTPException(r, 'failed to get image')
